from enum import Enum


class EventVisibility(Enum):
    DEFAULT = "default"
    PUBLIC = "public"
    PRIVATE = "private"
    CONFIDENTIAL = "confidential"


class SendUpdatesOptions(Enum):
    NONE = "none"  # No notifications are sent
    ALL = "all"  # Notifications are sent to all guests
    EXTERNAL_ONLY = "externalOnly"  # Notifications are sent to non-Google Calendar guests only.



from datetime import datetime
from unittest.mock import MagicMock, patch
from zoneinfo import ZoneInfo

import pytest
from arcade_tdk import ToolAuthorizationContext, ToolContext
from arcade_tdk.errors import RetryableToolError, ToolExecutionError
from googleapiclient.errors import HttpError

from arcade_google_calendar.enums import EventVisibility, SendUpdatesOptions
from arcade_google_calendar.tools import (
    create_event,
    delete_event,
    find_time_slots_when_everyone_is_free,
    list_calendars,
    list_events,
    update_event,
)


@pytest.fixture
def mock_context():
    mock_auth = ToolAuthorizationContext(token="fake-token")  # noqa: S106
    return ToolContext(authorization=mock_auth)


@pytest.mark.asyncio
@patch("arcade_google_calendar.tools.calendar.build_calendar_service")
async def test_list_calendars(mock_build_calendar_service, mock_context):
    mock_service = MagicMock()
    mock_build_calendar_service.return_value = mock_service

    expected_api_response = {
        "etag": '"p33for2n0pvc8o0o"',
        "items": [
            {
                "accessRole": "reader",
                "backgroundColor": "#16a765",
                "colorId": "8",
                "conferenceProperties": {"allowedConferenceSolutionTypes": ["hangoutsMeet"]},
                "defaultReminders": [],
                "description": "Holidays and Observances in Brazil",
                "etag": '"2347287866334000"',
                "foregroundColor": "#000000",
                "id": "en.brazilian#holiday@group.v.calendar.google.com",
                "kind": "calendar#calendarListEntry",
                "selected": True,
                "summary": "Holidays in Brazil",
                "timeZone": "America/Sao_Paulo",
            },
            {
                "accessRole": "owner",
                "backgroundColor": "#9fe1e7",
                "colorId": "14",
                "conferenceProperties": {"allowedConferenceSolutionTypes": ["hangoutsMeet"]},
                "defaultReminders": [{"method": "popup", "minutes": 10}],
                "etag": '"1743169667849567"',
                "foregroundColor": "#000000",
                "id": "example@arcade.dev",
                "kind": "calendar#calendarListEntry",
                "notificationSettings": {
                    "notifications": [
                        {"method": "email", "type": "eventCreation"},
                        {"method": "email", "type": "eventChange"},
                        {"method": "email", "type": "eventCancellation"},
                        {"method": "email", "type": "eventResponse"},
                    ]
                },
                "primary": True,
                "selected": True,
                "summary": "example@arcade.dev",
                "timeZone": "America/Sao_Paulo",
            },
        ],
        "kind": "calendar#calendarList",
        "nextSyncToken": "XkJ8Hy5mN2pQvL9sR4tW7cA3fE1iU6nB",
    }

    expected_tool_response = {
        "num_calendars": 2,
        "calendars": [
            {
                "description": "Holidays and Observances in Brazil",
                "id": "en.brazilian#holiday@group.v.calendar.google.com",
                "summary": "Holidays in Brazil",
                "timeZone": "America/Sao_Paulo",
            },
            {
                "id": "example@arcade.dev",
                "summary": "example@arcade.dev",
                "timeZone": "America/Sao_Paulo",
            },
        ],
        "next_page_token": None,
    }

    mock_service.calendarList().list().execute.return_value = expected_api_response

    response = await list_calendars(context=mock_context)
    assert response == expected_tool_response

    # Case: HttpError during calendars listing
    mock_service.calendarList().list().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Invalid request"}}',
    )

    with pytest.raises(ToolExecutionError):
        await list_calendars(context=mock_context)


@pytest.mark.asyncio
@patch("arcade_google_calendar.tools.calendar.build_calendar_service")
async def test_create_event(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Mock the calendar's time zone
    mock_service.calendars().get().execute.return_value = {"timeZone": "America/Los_Angeles"}

    # Case: HttpError during event creation
    mock_service.events().insert().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Invalid request"}}',
    )

    with pytest.raises(ToolExecutionError):
        await create_event(
            context=mock_context,
            summary="Test Event",
            start_datetime="2024-12-31T15:30:00",
            end_datetime="2024-12-31T17:30:00",
            description="Test Description",
            location="Test Location",
            visibility=EventVisibility.PRIVATE,
            attendee_emails=["test@example.com"],
        )


@pytest.mark.asyncio
@patch("arcade_google_calendar.tools.calendar.build_calendar_service")
async def test_list_events(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service
    # Mock the calendar's time zone
    mock_service.calendars().get().execute.return_value = {"timeZone": "America/Los_Angeles"}

    # Mock the events list response
    mock_events_list_response = {
        "items": [
            {
                "creator": {"email": "example@arcade-ai.com", "self": True},
                "end": {"dateTime": "2024-09-27T01:00:00-07:00", "timeZone": "America/Los_Angeles"},
                "eventType": "default",
                "htmlLink": "https://www.google.com/calendar/event?eid=event1",
                "id": "event1",
                "organizer": {"email": "example@arcade-ai.com", "self": True},
                "start": {
                    "dateTime": "2024-09-27T00:00:00-07:00",
                    "timeZone": "America/Los_Angeles",
                },
                "summary": "Event 1",
            },
            {
                "creator": {"email": "example@arcade-ai.com", "self": True},
                "end": {"dateTime": "2024-09-27T17:00:00-07:00", "timeZone": "America/Los_Angeles"},
                "eventType": "default",
                "htmlLink": "https://www.google.com/calendar/event?eid=event2",
                "id": "event2",
                "organizer": {"email": "example@arcade-ai.com", "self": True},
                "start": {
                    "dateTime": "2024-09-27T14:00:00-07:00",
                    "timeZone": "America/Los_Angeles",
                },
                "summary": "Event 2",
            },
        ]
    }
    expected_tool_response = {
        "events_count": len(mock_events_list_response["items"]),
        "events": mock_events_list_response["items"],
    }
    mock_service.events().list().execute.return_value = mock_events_list_response
    response = await list_events(
        context=mock_context,
        min_end_datetime="2024-09-15T09:00:00",
        max_start_datetime="2024-09-16T17:00:00",
    )
    assert response == expected_tool_response

    # Case: HttpError during events listing
    mock_service.events().list().execute.side_effect = HttpError(
        resp=MagicMock(status=400),
        content=b'{"error": {"message": "Invalid request"}}',
    )

    with pytest.raises(ToolExecutionError):
        await list_events(
            context=mock_context,
            min_end_datetime="2024-09-15T09:00:00",
            max_start_datetime="2024-09-16T17:00:00",
        )


@pytest.mark.asyncio
@patch("arcade_google_calendar.tools.calendar.build_calendar_service")
async def test_update_event(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service

    # Mock retrieval of the event
    mock_service.events().get().execute.side_effect = HttpError(
        resp=MagicMock(status=404),
        content=b'{"error": {"message": "Event not found"}}',
    )

    with pytest.raises(ToolExecutionError):
        await update_event(
            context=mock_context,
            event_id="1234567890",
            updated_start_datetime="2024-12-31T00:15:00",
            updated_end_datetime="2024-12-31T01:15:00",
            updated_summary="Updated Event",
            updated_description="Updated Description",
            updated_location="Updated Location",
            updated_visibility=EventVisibility.PRIVATE,
            attendee_emails_to_add=["test@example.com"],
            attendee_emails_to_remove=["test@example2.com"],
            send_updates=SendUpdatesOptions.ALL,
        )


@pytest.mark.asyncio
@patch("arcade_google_calendar.tools.calendar.build_calendar_service")
async def test_delete_event(mock_build, mock_context):
    mock_service = MagicMock()
    mock_build.return_value = mock_service
    mock_service.events().delete().execute.side_effect = HttpError(
        resp=MagicMock(status=404),
        content=b'{"error": {"message": "Event not found"}}',
    )

    with pytest.raises(ToolExecutionError):
        await delete_event(
            context=mock_context,
            event_id="nonexistent_event",
            send_updates=SendUpdatesOptions.ALL,
        )


@pytest.mark.asyncio
@patch("arcade_google_calendar.utils.get_now")
@patch("arcade_google_calendar.tools.calendar.build_oauth_service")
@patch("arcade_google_calendar.tools.calendar.build_calendar_service")
async def test_find_free_slots_happiest_path_single_user(
    mock_build_calendar_service, mock_build_oauth_service, mock_get_now, mock_context
):
    calendar_service = MagicMock()
    oauth_service = MagicMock()

    mock_get_now.return_value = datetime(
        2025, 3, 10, 9, 25, 0, tzinfo=ZoneInfo("America/Los_Angeles")
    )
    mock_build_oauth_service.return_value = oauth_service
    mock_build_calendar_service.return_value = calendar_service

    oauth_service.userinfo().get().execute.return_value = {
        "email": "example@arcade.dev",
    }

    calendar_service.freebusy().query().execute.return_value = {
        "calendars": {
            "example@arcade.dev": {"busy": []},
        }
    }

    calendar_service.calendars().get().execute.return_value = {
        "timeZone": "America/Los_Angeles",
    }

    response = await find_time_slots_when_everyone_is_free(
        context=mock_context,
        email_addresses=["example@arcade.dev"],
        start_date="2025-03-10",
        end_date="2025-03-11",
        start_time_boundary="08:00",
        end_time_boundary="18:00",
    )

    assert response == {
        "free_slots": [
            {
                "start": {
                    "datetime": "2025-03-10T09:25:00-07:00",
                    "weekday": "Monday",
                },
                "end": {
                    "datetime": "2025-03-10T18:00:00-07:00",
                    "weekday": "Monday",
                },
            },
            {
                "start": {
                    "datetime": "2025-03-11T08:00:00-07:00",
                    "weekday": "Tuesday",
                },
                "end": {
                    "datetime": "2025-03-11T18:00:00-07:00",
                    "weekday": "Tuesday",
                },
            },
        ],
        "timezone": "America/Los_Angeles",
    }


@pytest.mark.asyncio
@patch("arcade_google_calendar.utils.get_now")
@patch("arcade_google_calendar.tools.calendar.build_oauth_service")
@patch("arcade_google_calendar.tools.calendar.build_calendar_service")
async def test_find_free_slots_happiest_path_single_user_with_busy_times(
    mock_build_calendar_service, mock_build_oauth_service, mock_get_now, mock_context
):
    calendar_service = MagicMock()
    oauth_service = MagicMock()

    mock_get_now.return_value = datetime(
        2025, 3, 10, 9, 25, 0, tzinfo=ZoneInfo("America/Los_Angeles")
    )

    mock_build_oauth_service.return_value = oauth_service
    mock_build_calendar_service.return_value = calendar_service

    oauth_service.userinfo().get().execute.return_value = {
        "email": "example@arcade.dev",
    }

    calendar_service.freebusy().query().execute.return_value = {
        "calendars": {
            "example@arcade.dev": {
                "busy": [
                    {
                        "start": "2025-03-10T11:00:00-07:00",
                        "end": "2025-03-10T12:00:00-07:00",
                    },
                    {
                        "start": "2025-03-10T14:15:00-07:00",
                        "end": "2025-03-10T14:30:00-07:00",
                    },
                ]
            },
        }
    }

    calendar_service.calendars().get().execute.return_value = {
        "timeZone": "America/Los_Angeles",
    }

    response = await find_time_slots_when_everyone_is_free(
        context=mock_context,
        email_addresses=["example@arcade.dev"],
        start_date="2025-03-10",
        end_date="2025-03-11",
        start_time_boundary="08:00",
        end_time_boundary="18:00",
    )

    assert response == {
        "free_slots": [
            {
                "start": {
                    "datetime": "2025-03-10T09:25:00-07:00",
                    "weekday": "Monday",
                },
                "end": {
                    "datetime": "2025-03-10T11:00:00-07:00",
                    "weekday": "Monday",
                },
            },
            {
                "start": {
                    "datetime": "2025-03-10T12:00:00-07:00",
                    "weekday": "Monday",
                },
                "end": {
                    "datetime": "2025-03-10T14:15:00-07:00",
                    "weekday": "Monday",
                },
            },
            {
                "start": {
                    "datetime": "2025-03-10T14:30:00-07:00",
                    "weekday": "Monday",
                },
                "end": {
                    "datetime": "2025-03-10T18:00:00-07:00",
                    "weekday": "Monday",
                },
            },
            {
                "start": {
                    "datetime": "2025-03-11T08:00:00-07:00",
                    "weekday": "Tuesday",
                },
                "end": {
                    "datetime": "2025-03-11T18:00:00-07:00",
                    "weekday": "Tuesday",
                },
            },
        ],
        "timezone": "America/Los_Angeles",
    }


@pytest.mark.asyncio
@patch("arcade_google_calendar.utils.get_now")
@patch("arcade_google_calendar.tools.calendar.build_oauth_service")
@patch("arcade_google_calendar.tools.calendar.build_calendar_service")
async def test_find_free_slots_happiest_path_multiple_users_with_busy_times(
    mock_build_calendar_service, mock_build_oauth_service, mock_get_now, mock_context
):
    calendar_service = MagicMock()
    oauth_service = MagicMock()

    mock_get_now.return_value = datetime(
        2025, 3, 10, 9, 25, 0, tzinfo=ZoneInfo("America/Los_Angeles")
    )

    mock_build_oauth_service.return_value = oauth_service
    mock_build_calendar_service.return_value = calendar_service

    oauth_service.userinfo().get().execute.return_value = {
        "email": "example@arcade.dev",
    }

    calendar_service.freebusy().query().execute.return_value = {
        "calendars": {
            "example@arcade.dev": {
                "busy": [
                    {
                        "start": "2025-03-10T11:00:00-07:00",
                        "end": "2025-03-10T12:00:00-07:00",
                    },
                    {
                        "start": "2025-03-10T14:15:00-07:00",
                        "end": "2025-03-10T14:30:00-07:00",
                    },
                ]
            },
            "example2@arcade.dev": {
                "busy": [
                    {
                        "start": "2025-03-10T11:30:00-07:00",
                        "end": "2025-03-10T12:45:00-07:00",
                    },
                    {
                        "start": "2025-03-11T06:00:00-07:00",
                        "end": "2025-03-11T07:00:00-07:00",
                    },
                ]
            },
        }
    }

    calendar_service.calendars().get().execute.return_value = {
        "timeZone": "America/Los_Angeles",
    }

    response = await find_time_slots_when_everyone_is_free(
        context=mock_context,
        email_addresses=["example@arcade.dev", "example2@arcade.dev"],
        start_date="2025-03-10",
        end_date="2025-03-11",
        start_time_boundary="08:00",
        end_time_boundary="18:00",
    )

    assert response == {
        "free_slots": [
            {
                "start": {
                    "datetime": "2025-03-10T09:25:00-07:00",
                    "weekday": "Monday",
                },
                "end": {
                    "datetime": "2025-03-10T11:00:00-07:00",
                    "weekday": "Monday",
                },
            },
            {
                "start": {
                    "datetime": "2025-03-10T12:45:00-07:00",
                    "weekday": "Monday",
                },
                "end": {
                    "datetime": "2025-03-10T14:15:00-07:00",
                    "weekday": "Monday",
                },
            },
            {
                "start": {
                    "datetime": "2025-03-10T14:30:00-07:00",
                    "weekday": "Monday",
                },
                "end": {
                    "datetime": "2025-03-10T18:00:00-07:00",
                    "weekday": "Monday",
                },
            },
            {
                "start": {
                    "datetime": "2025-03-11T08:00:00-07:00",
                    "weekday": "Tuesday",
                },
                "end": {
                    "datetime": "2025-03-11T18:00:00-07:00",
                    "weekday": "Tuesday",
                },
            },
        ],
        "timezone": "America/Los_Angeles",
    }


@pytest.mark.asyncio
@patch("arcade_google_calendar.utils.get_now")
@patch("arcade_google_calendar.tools.calendar.build_oauth_service")
@patch("arcade_google_calendar.tools.calendar.build_calendar_service")
async def test_find_free_slots_with_google_calendar_error_not_found(
    mock_build_calendar_service, mock_build_oauth_service, mock_get_now, mock_context
):
    calendar_service = MagicMock()
    oauth_service = MagicMock()

    mock_get_now.return_value = datetime(
        2025, 3, 10, 9, 25, 0, tzinfo=ZoneInfo("America/Los_Angeles")
    )
    mock_build_oauth_service.return_value = oauth_service
    mock_build_calendar_service.return_value = calendar_service

    oauth_service.userinfo().get().execute.return_value = {
        "email": "example@arcade.dev",
    }

    calendar_service.freebusy().query().execute.return_value = {
        "calendars": {
            "example@arcade.dev": {
                "busy": [
                    {
                        "start": "2025-03-10T11:00:00-07:00",
                        "end": "2025-03-10T12:00:00-07:00",
                    },
                    {
                        "start": "2025-03-10T14:15:00-07:00",
                        "end": "2025-03-10T14:30:00-07:00",
                    },
                ]
            },
            "example2@arcade.dev": {
                "errors": [
                    {
                        "reason": "notFound",
                        "domain": "calendar",
                    }
                ]
            },
        }
    }

    calendar_service.calendars().get().execute.return_value = {
        "timeZone": "America/Los_Angeles",
    }

    with pytest.raises(RetryableToolError):
        await find_time_slots_when_everyone_is_free(
            context=mock_context,
            email_addresses=["example@arcade.dev", "example2@arcade.dev"],
            start_date="2025-03-10",
            end_date="2025-03-11",
            start_time_boundary="08:00",
            end_time_boundary="18:00",
        )



import logging
from datetime import date, datetime, time, timedelta, timezone
from typing import Any
from zoneinfo import ZoneInfo

from google.oauth2.credentials import Credentials
from googleapiclient.discovery import Resource, build

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)

logger = logging.getLogger(__name__)


def parse_datetime(datetime_str: str, time_zone: str) -> datetime:
    """
    Parse a datetime string in ISO 8601 format and ensure it is timezone-aware.

    Args:
        datetime_str (str): The datetime string to parse. Expected format: 'YYYY-MM-DDTHH:MM:SS'.
        time_zone (str): The timezone to apply if the datetime string is naive.

    Returns:
        datetime: A timezone-aware datetime object.

    Raises:
        ValueError: If the datetime string is not in the correct format.
    """
    datetime_str = datetime_str.upper().strip().rstrip("Z")
    try:
        dt = datetime.fromisoformat(datetime_str)
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=ZoneInfo(time_zone))
    except ValueError as e:
        raise ValueError(
            f"Invalid datetime format: '{datetime_str}'. "
            "Expected ISO 8601 format, e.g., '2024-12-31T15:30:00'."
        ) from e
    return dt


def build_oauth_service(auth_token: str | None) -> Resource:  # type: ignore[no-any-unimported]
    """
    Build an OAuth2 service object.
    """
    auth_token = auth_token or ""
    return build("oauth2", "v2", credentials=Credentials(auth_token))


def build_calendar_service(auth_token: str | None) -> Resource:  # type: ignore[no-any-unimported]
    """
    Build a Calendar service object.
    """
    auth_token = auth_token or ""
    return build("calendar", "v3", credentials=Credentials(auth_token))


def weekday_to_name(weekday: int) -> str:
    return ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"][weekday]


def get_time_boundaries_for_date(
    current_date: date,
    global_start: datetime,
    global_end: datetime,
    start_time_boundary: time,
    end_time_boundary: time,
    tz: ZoneInfo,
) -> tuple[datetime, datetime]:
    """Compute the allowed start and end times for the given day, adjusting for global bounds."""
    day_start_time = datetime.combine(current_date, start_time_boundary).replace(tzinfo=tz)
    day_end_time = datetime.combine(current_date, end_time_boundary).replace(tzinfo=tz)

    if current_date == global_start.date():
        day_start_time = max(day_start_time, global_start)

    if current_date == global_end.date():
        day_end_time = min(day_end_time, global_end)

    return day_start_time, day_end_time


def gather_busy_intervals(
    busy_data: dict[str, Any],
    day_start: datetime,
    day_end: datetime,
    business_tz: ZoneInfo,
) -> list[tuple[datetime, datetime]]:
    """
    Collect busy intervals from all calendars that intersect with the day's business hours.
    Busy intervals are clipped to lie within [day_start, day_end].
    """
    busy_intervals = []
    for calendar in busy_data:
        for slot in busy_data[calendar].get("busy", []):
            slot_start = parse_rfc3339_datetime_str(slot["start"]).astimezone(business_tz)
            slot_end = parse_rfc3339_datetime_str(slot["end"]).astimezone(business_tz)
            if slot_end > day_start and slot_start < day_end:
                busy_intervals.append((max(slot_start, day_start), min(slot_end, day_end)))
    return busy_intervals


def subtract_busy_intervals(
    business_start: datetime,
    business_end: datetime,
    busy_intervals: list[tuple[datetime, datetime]],
) -> list[dict[str, Any]]:
    """
    Subtract the merged busy intervals from the business hours and return free time slots.
    """
    free_slots = []
    merged_busy = merge_intervals(busy_intervals)

    # If there are no busy intervals, return the entire business window as free.
    if not merged_busy:
        return [
            {
                "start": {
                    "datetime": business_start.isoformat(),
                    "weekday": weekday_to_name(business_start.weekday()),
                },
                "end": {
                    "datetime": business_end.isoformat(),
                    "weekday": weekday_to_name(business_end.weekday()),
                },
            }
        ]

    current_free_start = business_start
    for busy_start, busy_end in merged_busy:
        if current_free_start < busy_start:
            free_slots.append({
                "start": {
                    "datetime": current_free_start.isoformat(),
                    "weekday": weekday_to_name(current_free_start.weekday()),
                },
                "end": {
                    "datetime": busy_start.isoformat(),
                    "weekday": weekday_to_name(busy_start.weekday()),
                },
            })
        current_free_start = max(current_free_start, busy_end)
    if current_free_start < business_end:
        free_slots.append({
            "start": {
                "datetime": current_free_start.isoformat(),
                "weekday": weekday_to_name(current_free_start.weekday()),
            },
            "end": {
                "datetime": business_end.isoformat(),
                "weekday": weekday_to_name(business_end.weekday()),
            },
        })
    return free_slots


def compute_free_time_intersection(
    busy_data: dict[str, Any],
    global_start: datetime,
    global_end: datetime,
    start_time_boundary: time,
    end_time_boundary: time,
    include_weekends: bool,
    tz: ZoneInfo,
) -> list[dict[str, Any]]:
    """
    Returns the free time slots across all calendars within the global bounds,
    ensuring that the global start is not in the past.

    Only considers business days (Monday to Friday) and business hours (08:00-19:00)
    in the provided timezone.
    """
    # Ensure global_start is never in the past relative to now.
    now = get_now(tz)

    if now > global_start:
        global_start = now

    # If after adjusting the start, there's no interval left, return empty.
    if global_start >= global_end:
        return []

    free_slots = []
    current_date = global_start.date()

    while current_date <= global_end.date():
        if not include_weekends and current_date.weekday() >= 5:
            current_date += timedelta(days=1)
            continue

        day_start, day_end = get_time_boundaries_for_date(
            current_date=current_date,
            global_start=global_start,
            global_end=global_end,
            start_time_boundary=start_time_boundary,
            end_time_boundary=end_time_boundary,
            tz=tz,
        )

        # Skip if the day's allowed time window is empty.
        if day_start >= day_end:
            current_date += timedelta(days=1)
            continue

        busy_intervals = gather_busy_intervals(busy_data, day_start, day_end, tz)
        free_slots.extend(subtract_busy_intervals(day_start, day_end, busy_intervals))

        current_date += timedelta(days=1)

    return free_slots


def get_now(tz: ZoneInfo | None = None) -> datetime:
    if not tz:
        tz = ZoneInfo("UTC")
    return datetime.now(tz)


def parse_rfc3339_datetime_str(dt_str: str, tz: timezone = timezone.utc) -> datetime:
    """
    Parse an RFC3339 datetime string into a timezone-aware datetime.
    Converts a trailing 'Z' (UTC) into +00:00.
    If the parsed datetime is naive, assume it is in the provided timezone.
    """
    if dt_str.endswith("Z"):
        dt_str = dt_str[:-1] + "+00:00"
    dt = datetime.fromisoformat(dt_str)
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=tz)
    return dt


def merge_intervals(intervals: list[tuple[datetime, datetime]]) -> list[tuple[datetime, datetime]]:
    """
    Given a list of (start, end) tuples, merge overlapping or adjacent intervals.
    """
    merged: list[tuple[datetime, datetime]] = []
    for start, end in sorted(intervals, key=lambda x: x[0]):
        if not merged:
            merged.append((start, end))
        else:
            last_start, last_end = merged[-1]
            if start <= last_end:
                merged[-1] = (last_start, max(last_end, end))
            else:
                merged.append((start, end))
    return merged
